/* Copyright (c) 2004-2006, Sara Golemon <sarag@libssh2.org>
 * Copyright (c) 2009-2014 by Daniel Stenberg
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *
 *   Redistributions of source code must retain the above
 *   copyright notice, this list of conditions and the
 *   following disclaimer.
 *
 *   Redistributions in binary form must reproduce the above
 *   copyright notice, this list of conditions and the following
 *   disclaimer in the documentation and/or other materials
 *   provided with the distribution.
 *
 *   Neither the name of the copyright holder nor the names
 *   of any other contributors may be used to endorse or
 *   promote products derived from this software without
 *   specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 */

#include "libssh2_priv.h"
#include "misc.h"
#include "ed25519.h"

/* Needed for struct iovec on some platforms */
#ifdef HAVE_SYS_UIO_H
#include <sys/uio.h>
#endif

#if LIBSSH2_RSA
/* ***********
 * ssh-rsa *
 *********** */

static int hostkey_method_ssh_rsa_dtor(LIBSSH2_SESSION * session,
                                       void **abstract);

/*
 * hostkey_method_ssh_rsa_init
 *
 * Initialize the server hostkey working area with e/n pair
 */
static int
hostkey_method_ssh_rsa_init(LIBSSH2_SESSION * session,
                            const unsigned char *hostkey_data,
                            size_t hostkey_data_len,
                            void **abstract)
{
    libssh2_rsa_ctx *rsactx;
    const unsigned char *s, *e, *n;
    unsigned long len, e_len, n_len;
    int ret;

    (void) hostkey_data_len;

    if (*abstract) {
        hostkey_method_ssh_rsa_dtor(session, abstract);
        *abstract = NULL;
    }

    s = hostkey_data;
    len = _libssh2_ntohu32(s);
    s += 4;

    if (len != 7 || strncmp((char *) s, "ssh-rsa", 7) != 0) {
        return -1;
    }
    s += 7;

    e_len = _libssh2_ntohu32(s);
    s += 4;

    e = s;
    s += e_len;
    n_len = _libssh2_ntohu32(s);
    s += 4;
    n = s;

    ret = _libssh2_rsa_new(&rsactx, e, e_len, n, n_len, NULL, 0,
                           NULL, 0, NULL, 0, NULL, 0, NULL, 0, NULL, 0);
    if (ret) {
        return -1;
    }

    *abstract = rsactx;

    return 0;
}

/*
 * hostkey_method_ssh_rsa_initPEM
 *
 * Load a Private Key from a PEM file
 */
static int
hostkey_method_ssh_rsa_initPEM(LIBSSH2_SESSION * session,
                               const char *privkeyfile,
                               unsigned const char *passphrase,
                               void **abstract)
{
    libssh2_rsa_ctx *rsactx;
    int ret;

    if (*abstract) {
        hostkey_method_ssh_rsa_dtor(session, abstract);
        *abstract = NULL;
    }

    ret = _libssh2_rsa_new_private(&rsactx, session, privkeyfile, passphrase);
    if (ret) {
        return -1;
    }

    *abstract = rsactx;

    return 0;
}

/*
 * hostkey_method_ssh_rsa_initPEMFromMemory
 *
 * Load a Private Key from a memory
 */
static int
hostkey_method_ssh_rsa_initPEMFromMemory(LIBSSH2_SESSION * session,
                                         const char *privkeyfiledata,
                                         size_t privkeyfiledata_len,
                                         unsigned const char *passphrase,
                                         void **abstract)
{
    libssh2_rsa_ctx *rsactx;
    int ret;

    if (*abstract) {
        hostkey_method_ssh_rsa_dtor(session, abstract);
        *abstract = NULL;
    }

    ret = _libssh2_rsa_new_private_frommemory(&rsactx, session,
                                              privkeyfiledata,
                                              privkeyfiledata_len, passphrase);
    if (ret) {
        return -1;
    }

    *abstract = rsactx;

    return 0;
}

/*
 * hostkey_method_ssh_rsa_sign
 *
 * Verify signature created by remote
 */
static int
hostkey_method_ssh_rsa_sig_verify(LIBSSH2_SESSION * session,
                                  const unsigned char *sig,
                                  size_t sig_len,
                                  const unsigned char *m,
                                  size_t m_len, void **abstract)
{
    libssh2_rsa_ctx *rsactx = (libssh2_rsa_ctx *) (*abstract);
    (void) session;

    /* Skip past keyname_len(4) + keyname(7){"ssh-rsa"} + signature_len(4) */
    sig += 15;
    sig_len -= 15;
    return _libssh2_rsa_sha1_verify(rsactx, sig, sig_len, m, m_len);
}

/*
 * hostkey_method_ssh_rsa_signv
 *
 * Construct a signature from an array of vectors
 */
static int
hostkey_method_ssh_rsa_signv(LIBSSH2_SESSION * session,
                             unsigned char **signature,
                             size_t *signature_len,
                             int veccount,
                             const struct iovec datavec[],
                             void **abstract)
{
    libssh2_rsa_ctx *rsactx = (libssh2_rsa_ctx *) (*abstract);

#ifdef _libssh2_rsa_sha1_signv
    return _libssh2_rsa_sha1_signv(session, signature, signature_len,
                                   veccount, datavec, rsactx);
#else
    int ret;
    int i;
    unsigned char hash[SHA_DIGEST_LENGTH];
    libssh2_sha1_ctx ctx;

    libssh2_sha1_init(&ctx);
    for(i = 0; i < veccount; i++) {
        libssh2_sha1_update(ctx, datavec[i].iov_base, datavec[i].iov_len);
    }
    libssh2_sha1_final(ctx, hash);

    ret = _libssh2_rsa_sha1_sign(session, rsactx, hash, SHA_DIGEST_LENGTH,
                                 signature, signature_len);
    if (ret) {
        return -1;
    }

    return 0;
#endif
}

/*
 * hostkey_method_ssh_rsa_dtor
 *
 * Shutdown the hostkey
 */
static int
hostkey_method_ssh_rsa_dtor(LIBSSH2_SESSION * session, void **abstract)
{
    libssh2_rsa_ctx *rsactx = (libssh2_rsa_ctx *) (*abstract);
    (void) session;

    _libssh2_rsa_free(rsactx);

    *abstract = NULL;

    return 0;
}

#ifdef OPENSSL_NO_MD5
#define MD5_DIGEST_LENGTH 16
#endif

static const LIBSSH2_HOSTKEY_METHOD hostkey_method_ssh_rsa = {
    "ssh-rsa",
    MD5_DIGEST_LENGTH,
    hostkey_method_ssh_rsa_init,
    hostkey_method_ssh_rsa_initPEM,
    hostkey_method_ssh_rsa_initPEMFromMemory,
    hostkey_method_ssh_rsa_sig_verify,
    hostkey_method_ssh_rsa_signv,
    NULL,                       /* encrypt */
    hostkey_method_ssh_rsa_dtor,
};
#endif /* LIBSSH2_RSA */

#if LIBSSH2_DSA
/* ***********
 * ssh-dss *
 *********** */

static int hostkey_method_ssh_dss_dtor(LIBSSH2_SESSION * session,
                                       void **abstract);

/*
 * hostkey_method_ssh_dss_init
 *
 * Initialize the server hostkey working area with p/q/g/y set
 */
static int
hostkey_method_ssh_dss_init(LIBSSH2_SESSION * session,
                            const unsigned char *hostkey_data,
                            size_t hostkey_data_len,
                            void **abstract)
{
    libssh2_dsa_ctx *dsactx;
    const unsigned char *p, *q, *g, *y, *s;
    unsigned long p_len, q_len, g_len, y_len, len;
    int ret;

    (void) hostkey_data_len;

    if (*abstract) {
        hostkey_method_ssh_dss_dtor(session, abstract);
        *abstract = NULL;
    }

    s = hostkey_data;
    len = _libssh2_ntohu32(s);
    s += 4;
    if (len != 7 || strncmp((char *) s, "ssh-dss", 7) != 0) {
        return -1;
    }
    s += 7;

    p_len = _libssh2_ntohu32(s);
    s += 4;
    p = s;
    s += p_len;
    q_len = _libssh2_ntohu32(s);
    s += 4;
    q = s;
    s += q_len;
    g_len = _libssh2_ntohu32(s);
    s += 4;
    g = s;
    s += g_len;
    y_len = _libssh2_ntohu32(s);
    s += 4;
    y = s;
    /* s += y_len; */

    ret = _libssh2_dsa_new(&dsactx, p, p_len, q, q_len,
                           g, g_len, y, y_len, NULL, 0);
    if (ret) {
        return -1;
    }

    *abstract = dsactx;

    return 0;
}

/*
 * hostkey_method_ssh_dss_initPEM
 *
 * Load a Private Key from a PEM file
 */
static int
hostkey_method_ssh_dss_initPEM(LIBSSH2_SESSION * session,
                               const char *privkeyfile,
                               unsigned const char *passphrase,
                               void **abstract)
{
    libssh2_dsa_ctx *dsactx;
    int ret;

    if (*abstract) {
        hostkey_method_ssh_dss_dtor(session, abstract);
        *abstract = NULL;
    }

    ret = _libssh2_dsa_new_private(&dsactx, session, privkeyfile, passphrase);
    if (ret) {
        return -1;
    }

    *abstract = dsactx;

    return 0;
}

/*
 * hostkey_method_ssh_dss_initPEMFromMemory
 *
 * Load a Private Key from memory
 */
static int
hostkey_method_ssh_dss_initPEMFromMemory(LIBSSH2_SESSION * session,
                                         const char *privkeyfiledata,
                                         size_t privkeyfiledata_len,
                                         unsigned const char *passphrase,
                                         void **abstract)
{
    libssh2_dsa_ctx *dsactx;
    int ret;

    if (*abstract) {
        hostkey_method_ssh_dss_dtor(session, abstract);
        *abstract = NULL;
    }

    ret = _libssh2_dsa_new_private_frommemory(&dsactx, session,
                                              privkeyfiledata,
                                              privkeyfiledata_len, passphrase);
    if (ret) {
        return -1;
    }

    *abstract = dsactx;

    return 0;
}

/*
 * libssh2_hostkey_method_ssh_dss_sign
 *
 * Verify signature created by remote
 */
static int
hostkey_method_ssh_dss_sig_verify(LIBSSH2_SESSION * session,
                                  const unsigned char *sig,
                                  size_t sig_len,
                                  const unsigned char *m,
                                  size_t m_len, void **abstract)
{
    libssh2_dsa_ctx *dsactx = (libssh2_dsa_ctx *) (*abstract);

    /* Skip past keyname_len(4) + keyname(7){"ssh-dss"} + signature_len(4) */
    sig += 15;
    sig_len -= 15;
    if (sig_len != 40) {
        return _libssh2_error(session, LIBSSH2_ERROR_PROTO,
                              "Invalid DSS signature length");
    }
    return _libssh2_dsa_sha1_verify(dsactx, sig, m, m_len);
}

/*
 * hostkey_method_ssh_dss_signv
 *
 * Construct a signature from an array of vectors
 */
static int
hostkey_method_ssh_dss_signv(LIBSSH2_SESSION * session,
                             unsigned char **signature,
                             size_t *signature_len,
                             int veccount,
                             const struct iovec datavec[],
                             void **abstract)
{
    libssh2_dsa_ctx *dsactx = (libssh2_dsa_ctx *) (*abstract);
    unsigned char hash[SHA_DIGEST_LENGTH];
    libssh2_sha1_ctx ctx;
    int i;

    *signature = LIBSSH2_CALLOC(session, 2 * SHA_DIGEST_LENGTH);
    if (!*signature) {
        return -1;
    }

    *signature_len = 2 * SHA_DIGEST_LENGTH;

    libssh2_sha1_init(&ctx);
    for(i = 0; i < veccount; i++) {
        libssh2_sha1_update(ctx, datavec[i].iov_base, datavec[i].iov_len);
    }
    libssh2_sha1_final(ctx, hash);

    if (_libssh2_dsa_sha1_sign(dsactx, hash, SHA_DIGEST_LENGTH, *signature)) {
        LIBSSH2_FREE(session, *signature);
        return -1;
    }

    return 0;
}

/*
 * libssh2_hostkey_method_ssh_dss_dtor
 *
 * Shutdown the hostkey method
 */
static int
hostkey_method_ssh_dss_dtor(LIBSSH2_SESSION * session, void **abstract)
{
    libssh2_dsa_ctx *dsactx = (libssh2_dsa_ctx *) (*abstract);
    (void) session;

    _libssh2_dsa_free(dsactx);

    *abstract = NULL;

    return 0;
}

static const LIBSSH2_HOSTKEY_METHOD hostkey_method_ssh_dss = {
    "ssh-dss",
    MD5_DIGEST_LENGTH,
    hostkey_method_ssh_dss_init,
    hostkey_method_ssh_dss_initPEM,
    hostkey_method_ssh_dss_initPEMFromMemory,
    hostkey_method_ssh_dss_sig_verify,
    hostkey_method_ssh_dss_signv,
    NULL,                       /* encrypt */
    hostkey_method_ssh_dss_dtor,
};
#endif /* LIBSSH2_DSA */








#if LIBSSH2_ECDSA
/* ******************************
 * ecdsa-sha2-nistp256-cert-v01 *
 ****************************** */

static int hostkey_method_ssh_ecdsa_dtor(LIBSSH2_SESSION * session,
                                       void **abstract);

/*
 * hostkey_method_ssh_dss_init
 *
 * Initialize the server hostkey working area with p/q/g/y set
 */
static int
hostkey_method_ssh_ecdsa_init(LIBSSH2_SESSION * session,
                            const unsigned char *hostkey_data,
                            size_t hostkey_data_len,
                            void **abstract)
{
    libssh2_ecdsa_ctx *ecdsactx;
    const unsigned char *s, *x, *y, *d;
    const char *s2;
    unsigned long xy_len, x_len, y_len, d_len, len, curvelen;
    int ret;
    unsigned int bitlen;

    if (*abstract) {
        hostkey_method_ssh_ecdsa_dtor(session, abstract);
        *abstract = NULL;
    }

    s = hostkey_data;
    s2 = (char*) s;
    len = _libssh2_ntohu32(s);
    s += 4;
    if (len < 19 || strncmp((char *) s, "ecdsa-sha2-nistp", 16) != 0) {
        return -1;
    }
    bitlen=(unsigned int)atoi((char*)s+16);
    s += 19;
    hostkey_data_len-=(19+4);
    curvelen=_libssh2_ntohu32(s);
    if (curvelen+4>hostkey_data_len)
        return -1;
    s += 4+curvelen;
    hostkey_data_len-=4+curvelen;

    xy_len = _libssh2_ntohu32(s);
    if (xy_len+4>hostkey_data_len || (xy_len&1)==0)
        return -1;
    x_len=(xy_len-1)/2;
    y_len=x_len;
    s += 5;
    x = s;
    s += x_len;
    y = s;
    s += y_len;
    hostkey_data_len-=4+xy_len;
    d_len = hostkey_data_len;    // just the remaining buffer!
    d = s;

    ret = _libssh2_ecdsa_new(&ecdsactx, bitlen, x, x_len, y, y_len, d, d_len);
    if (ret) {
        return -1;
    }

    *abstract = ecdsactx;

    return 0;
}

/*
 * hostkey_method_ssh_dss_initPEM
 *
 * Load a Private Key from a PEM file
 */
static int
hostkey_method_ssh_ecdsa_initPEM(LIBSSH2_SESSION * session,
                               const char *privkeyfile,
                               unsigned const char *passphrase,
                               void **abstract)
{
    libssh2_ecdsa_ctx *ecdsactx;
    int ret;

    if (*abstract) {
        hostkey_method_ssh_ecdsa_dtor(session, abstract);
        *abstract = NULL;
    }

    ret = _libssh2_ecdsa_new_private(&ecdsactx, session, privkeyfile, passphrase);
    if (ret) {
        return -1;
    }

    *abstract = ecdsactx;

    return 0;
}

/*
 * hostkey_method_ssh_dss_initPEMFromMemory
 *
 * Load a Private Key from memory
 */
static int
hostkey_method_ssh_ecdsa_initPEMFromMemory(LIBSSH2_SESSION * session,
                                         const char *privkeyfiledata,
                                         size_t privkeyfiledata_len,
                                         unsigned const char *passphrase,
                                         void **abstract)
{
    libssh2_ecdsa_ctx *ecdsactx;
    int ret;

    if (*abstract) {
        hostkey_method_ssh_ecdsa_dtor(session, abstract);
        *abstract = NULL;
    }

    ret = _libssh2_ecdsa_new_private_frommemory(&ecdsactx, session,
                                              privkeyfiledata,
                                              privkeyfiledata_len, passphrase);
    if (ret) {
        return -1;
    }

    *abstract = ecdsactx;

    return 0;
}

/*
 * libssh2_hostkey_method_ssh_dss_sign
 *
 * Verify signature created by remote
 */
static int
hostkey_method_ssh_ecdsa_sig_verify(LIBSSH2_SESSION * session,
                                  const unsigned char *sig,
                                  size_t sig_len,
                                  const unsigned char *m,
                                  size_t m_len, void **abstract)
{
    size_t name_len,data_len;
    libssh2_ecdsa_ctx *ecdsactx = (libssh2_ecdsa_ctx *) (*abstract);

    /* Skip past keyname_len(4) + keyname(19){"ecdsa-sha2-nistp256"} + signature_len(4) */
    name_len = _libssh2_ntohu32(sig);
    sig+=4;
    if (4+name_len+4>sig_len)
        return -1;
    sig+=name_len;
    sig_len-=4+name_len;
    data_len = _libssh2_ntohu32(sig);
    if (4+data_len!=sig_len || sig_len<64) {
        return _libssh2_error(session, LIBSSH2_ERROR_PROTO,
                              "Invalid DSS signature length");
    }
    sig+=4;
    sig_len-=4;
    return _libssh2_ecdsa_sha2_verify(ecdsactx, sig, sig_len, m, m_len);
}

/*
 * hostkey_method_ssh_dss_signv
 *
 * Construct a signature from an array of vectors
 */
static int
hostkey_method_ssh_ecdsa_signv(LIBSSH2_SESSION * session,
                             unsigned char **signature,
                             size_t *signature_len,
                             int veccount,
                             const struct iovec datavec[],
                             void **abstract)
{
    libssh2_ecdsa_ctx *ecdsactx = (libssh2_ecdsa_ctx *) (*abstract);
    unsigned char hash[SHA512_DIGEST_LENGTH];
    libssh2_sha256_ctx ctx256;
    libssh2_sha384_ctx ctx384;
    libssh2_sha512_ctx ctx512;
    int i;
    size_t hashlen;
    size_t signature_alloc_len=132+2+4+4;  // maximum possible length for ecdsa521, will resize below

    *signature = LIBSSH2_CALLOC(session, signature_alloc_len);
    if (!*signature) {
        return -1;
    }
    *signature_len = signature_alloc_len;

    switch (ecdsactx->cbEccKeyLength) {
	case 521:  // 521, not 512!
        hashlen = SHA512_DIGEST_LENGTH;
        libssh2_sha512_init(&ctx512);
        for(i = 0; i < veccount; i++) {
            libssh2_sha512_update(ctx512, datavec[i].iov_base, datavec[i].iov_len);
        }
        libssh2_sha512_final(ctx512, hash);
        break;
    case 384:
        hashlen = SHA384_DIGEST_LENGTH;
        libssh2_sha384_init(&ctx384);
        for(i = 0; i < veccount; i++) {
            libssh2_sha384_update(ctx384, datavec[i].iov_base, datavec[i].iov_len);
        }
        libssh2_sha384_final(ctx384, hash);
        break;
    default:
        hashlen = SHA256_DIGEST_LENGTH;
        libssh2_sha256_init(&ctx256);
        for(i = 0; i < veccount; i++) {
            libssh2_sha256_update(ctx256, datavec[i].iov_base, datavec[i].iov_len);
        }
        libssh2_sha256_final(ctx256, hash);
        break;
    }

    if (_libssh2_ecdsa_sha2_sign(ecdsactx, hash, hashlen, *signature, signature_len)) {
        LIBSSH2_FREE(session, *signature);
        return -1;
    }
    if (*signature_len<signature_alloc_len)
        LIBSSH2_REALLOC(session, *signature, *signature_len);
    return 0;
}

void
_libssh2_ecdsa_free(libssh2_ecdsa_ctx *ecdsa);

/*
 * libssh2_hostkey_method_ssh_dss_dtor
 *
 * Shutdown the hostkey method
 */
static int
hostkey_method_ssh_ecdsa_dtor(LIBSSH2_SESSION * session, void **abstract)
{
    libssh2_ecdsa_ctx *ecdsactx = (libssh2_ecdsa_ctx *) (*abstract);
    (void) session;

    _libssh2_ecdsa_free(ecdsactx);

    *abstract = NULL;

    return 0;
}

/*************************************************************************************/
static int
hostkey_method_ssh_ed25519_init(LIBSSH2_SESSION * session,
	const unsigned char *hostkey_data,
	size_t hostkey_data_len,
	void **abstract)
{
	unsigned char* key;
	unsigned char* s;
	int len;
	if (*abstract) {
		hostkey_method_ssh_ed25519_dtor(session, abstract);
		*abstract = NULL;
	}
	s = hostkey_data;
	len = _libssh2_ntohu32(s);
	s += 4;
	if (len != 11 || strncmp((char *)s, "ssh-ed25519", 11) != 0)
		return -1;
	hostkey_data += (4 + 11);
	hostkey_data_len -= (4 + 11);
	len = _libssh2_ntohu32(hostkey_data);
	if (len != 32)
		return -1;
	key = malloc(96);   // always allocate the full 96 bytes, so we can zero them when deleting!
	if (!key)
		return -1;
	memset(key, 0, 96);
	hostkey_data += 4;
	memcpy(key, hostkey_data, len);
	*abstract = key;
	return 0;
}

#define PEM_OPENSSH_HEADER "-----BEGIN OPENSSH PRIVATE KEY-----"
#define PEM_OPENSSH_FOOTER "-----END OPENSSH PRIVATE KEY-----"

static int
hostkey_method_ssh_ed25519_initPEM(LIBSSH2_SESSION * session,
	const char *privkeyfile,
	unsigned const char *passphrase,
	void **abstract)
{
	unsigned char *data = NULL;
	unsigned int datalen = 0;
	unsigned char *pbEncoded;
	unsigned long cbEncoded;
	unsigned char *key;
	unsigned char *pubkey;
	unsigned char *privkey;
	char* keyType;
	int keyTypeLen, pubkeyLen, privkeyLen;
	libssh2_ecdsa_ctx *ecdsactx;
	FILE *fp;
	char* cryptodata = NULL;
	int ret;

	if (*abstract) {
		hostkey_method_ssh_ecdsa_dtor(session, abstract);
		*abstract = NULL;
	}

	fp = fopen(privkeyfile, "r");
	if (!fp)
		return -1;

	ret = _libssh2_pem_parse(session, PEM_OPENSSH_HEADER, PEM_OPENSSH_FOOTER,
		fp, &data, &datalen, &cryptodata);  // cryptodata not used here, it's for the old private key format with weak encryption
	fclose(fp);
	if (ret)
		return -1;

	// openssh-key-v1 format?
	pbEncoded = data;
	cbEncoded = datalen;
	if (pbEncoded && strncmp("openssh-key-v1", pbEncoded, 15) == 0) {
		ret = _libssh2_pem_decrypt_bcrypt(session, passphrase,	&data, &datalen);
		if (ret == 0) {
			if (_libssh2_pem_read32bit(&pbEncoded, &cbEncoded) != _libssh2_pem_read32bit(&pbEncoded, &cbEncoded)) // verifyer
				return -1;
			if (_libssh2_pem_getByteArrayOffset(&pbEncoded, &cbEncoded, &keyType, &keyTypeLen) != 0)
				return -1;
			if (strnicmp(keyType, "ssh-ed25519", 11) == 0) {
				if (_libssh2_pem_getByteArrayOffset(&pbEncoded, &cbEncoded, &pubkey, &pubkeyLen) != 0)  // public key
					return -1;
				if (_libssh2_pem_getByteArrayOffset(&pbEncoded, &cbEncoded, &privkey, &privkeyLen) != 0)
					return -1;
				if (privkey == NULL || privkeyLen != 64)
					return -1;
				key = malloc(96);
				// Note: the ed25519 library uses the hashed seed (64 bytes) as "private key", not the seed sent by the server!
				ed25519_create_keypair(key, key + 32, privkey);
				*abstract = key;
				return 0;
			}
		}
	}
	return -1;
}

static int
hostkey_method_ssh_ed25519_sig_verify(LIBSSH2_SESSION * session,
	const unsigned char *sig,
	size_t sig_len,
	const unsigned char *m,
	size_t m_len, void **abstract)
{
	int len;
	len = _libssh2_ntohu32(sig);
	sig += 4;
	if (len != 11 || strncmp((char *)sig, "ssh-ed25519", 11) != 0)
		return -1;
	sig += 11;
	len = _libssh2_ntohu32(sig);
	sig += 4;
	if (len != 64)
		return -1;
	if (ed25519_verify(sig, m, m_len, *abstract))
		return 0;
	return -1;
}

static int
hostkey_method_ssh_ed25519_signv(LIBSSH2_SESSION * session,
	unsigned char **signature,
	size_t *signature_len,
	int veccount,
	const struct iovec datavec[],
	void **abstract)
{
	unsigned char* key = (unsigned char*)(*abstract);
	size_t sig_len = 64;

	*signature = LIBSSH2_CALLOC(session, sig_len);
	if (!*signature)
		return -1;
	*signature_len = sig_len;
	// ed25519 hashes the message by itself!
	ed25519_sign(*signature, datavec[0].iov_base, datavec[0].iov_len, key, key + 32);
	return 0;
}

static int
hostkey_method_ssh_ed25519_dtor(LIBSSH2_SESSION * session, void **abstract)
{
	if (*abstract) {
		memset(*abstract, 0, 96);
		free(*abstract);
		*abstract = NULL;
	}
	return 0;
}

/*************************************************************************************/

static const LIBSSH2_HOSTKEY_METHOD hostkey_method_ssh_ecdsa256 = {
    "ecdsa-sha2-nistp256",
    SHA256_DIGEST_LENGTH,
    hostkey_method_ssh_ecdsa_init,
    hostkey_method_ssh_ecdsa_initPEM,
    hostkey_method_ssh_ecdsa_initPEMFromMemory,
    hostkey_method_ssh_ecdsa_sig_verify,
    hostkey_method_ssh_ecdsa_signv,
    NULL,                       /* encrypt */
    hostkey_method_ssh_ecdsa_dtor,
};

static const LIBSSH2_HOSTKEY_METHOD hostkey_method_ssh_ecdsa384 = {
    "ecdsa-sha2-nistp384",
    SHA256_DIGEST_LENGTH,
    hostkey_method_ssh_ecdsa_init,
    hostkey_method_ssh_ecdsa_initPEM,
    hostkey_method_ssh_ecdsa_initPEMFromMemory,
    hostkey_method_ssh_ecdsa_sig_verify,
    hostkey_method_ssh_ecdsa_signv,
    NULL,                       /* encrypt */
    hostkey_method_ssh_ecdsa_dtor,
};

static const LIBSSH2_HOSTKEY_METHOD hostkey_method_ssh_ecdsa521 = {
    "ecdsa-sha2-nistp521",
    SHA256_DIGEST_LENGTH,
	hostkey_method_ssh_ecdsa_init,
	hostkey_method_ssh_ecdsa_initPEM,
	hostkey_method_ssh_ecdsa_initPEMFromMemory,
	hostkey_method_ssh_ecdsa_sig_verify,
	hostkey_method_ssh_ecdsa_signv,
	NULL,                       /* encrypt */
	hostkey_method_ssh_ecdsa_dtor,
};

static const LIBSSH2_HOSTKEY_METHOD hostkey_method_ssh_ed25519 = {
	"ssh-ed25519",
	SHA512_DIGEST_LENGTH,
	hostkey_method_ssh_ed25519_init,
	hostkey_method_ssh_ed25519_initPEM,
	NULL,
	hostkey_method_ssh_ed25519_sig_verify,
	hostkey_method_ssh_ed25519_signv,
	NULL,                       /* encrypt */
	hostkey_method_ssh_ed25519_dtor,
};
#endif /* LIBSSH2_ECDSA */


static const LIBSSH2_HOSTKEY_METHOD *hostkey_methods[] = {
#if LIBSSH2_ECDSA
    &hostkey_method_ssh_ecdsa521,
    &hostkey_method_ssh_ecdsa384,
    &hostkey_method_ssh_ecdsa256,
	&hostkey_method_ssh_ed25519,
#endif
/* LIBSSH2_DSA */
#if LIBSSH2_RSA
    &hostkey_method_ssh_rsa,
#endif /* LIBSSH2_RSA */
#if LIBSSH2_DSA
    &hostkey_method_ssh_dss,
#endif /* LIBSSH2_DSA */
    NULL
};

const LIBSSH2_HOSTKEY_METHOD **
libssh2_hostkey_methods(void)
{
    return hostkey_methods;
}

/*
 * libssh2_hostkey_hash
 *
 * Returns hash signature
 * Returned buffer should NOT be freed
 * Length of buffer is determined by hash type
 * i.e. MD5 == 16, SHA1 == 20
 */
LIBSSH2_API const char *
libssh2_hostkey_hash(LIBSSH2_SESSION * session, int hash_type)
{
    switch (hash_type) {
#if LIBSSH2_MD5
    case LIBSSH2_HOSTKEY_HASH_MD5:
        return (session->server_hostkey_md5_valid)
          ? (char *) session->server_hostkey_md5
          : NULL;
        break;
#endif /* LIBSSH2_MD5 */
    case LIBSSH2_HOSTKEY_HASH_SHA1:
        return (session->server_hostkey_sha1_valid)
          ? (char *) session->server_hostkey_sha1
          : NULL;
        break;
    default:
        return NULL;
    }
}

static int hostkey_type(const unsigned char *hostkey, size_t len)
{
    const unsigned char rsa[] = {
        0, 0, 0, 0x07, 's', 's', 'h', '-', 'r', 's', 'a'
    };
    const unsigned char dss[] = {
        0, 0, 0, 0x07, 's', 's', 'h', '-', 'd', 's', 's'
    };
    
    if (len < 11)
        return LIBSSH2_HOSTKEY_TYPE_UNKNOWN;

    if (!memcmp(rsa, hostkey, 11))
        return LIBSSH2_HOSTKEY_TYPE_RSA;

    if (!memcmp(dss, hostkey, 11))
        return LIBSSH2_HOSTKEY_TYPE_DSS;

    return LIBSSH2_HOSTKEY_TYPE_UNKNOWN;
}

/*
 * libssh2_session_hostkey()
 *
 * Returns the server key and length.
 *
 */
LIBSSH2_API const char *
libssh2_session_hostkey(LIBSSH2_SESSION *session, size_t *len, int *type)
{
    if(session->server_hostkey_len) {
        if(len)
            *len = session->server_hostkey_len;
        if (type)
            *type = hostkey_type(session->server_hostkey,
                                 session->server_hostkey_len);
        return (char *) session->server_hostkey;
    }
    if(len)
        *len = 0;
    return NULL;
}

